package edu.upvehu.gbg.tap.babel;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringJoiner;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 *
 * @author clase
 */
class Babel {
    Map<String, Language> lenguajes=new HashMap<>();

    /**
     * Constructor a parte del fichero "languages.txt" a encontrar en el directorio que se pasa como parámetro.
     * En ese mismo directorio debe encontrarse un subdirectorio "txt" con todos los ficheros de texto asociados a los
     * identificadores que aparecen en el fichero "languages.txt".<br>
     * 
     * En caso de no poder leer un fichero de algún lenguaje, se ignora e incluye un mensaje en el logger. 
     * 
     * @param dir directorio donde se encuentra la estructura de datos
     * @throws IOException Hay algún problema con la lectura del fichero "languages.txt" 
     */
    Babel(String dir) throws IOException {
        try(BufferedReader br=new BufferedReader(new FileReader(dir+"/languages.txt"))){
            String ln;
            while ((ln=br.readLine())!=null) {
                if (ln.trim().length()==0) continue;
                String[]  campos=ln.split("\\s+",2);
                try {
                    lenguajes.put(campos[0],new Language(campos[0],campos[1],dir+"/txt/"+campos[0]+".txt"));
                } catch (IOException ex) {
                    Logger.getLogger(Babel.class.getName()).log(Level.SEVERE, "No se hapodido leer el fichero del lenguaje "+campos[1]);
                }
            }
        }
    }   
        
    @Override
    public String toString() {
        StringJoiner sj=new StringJoiner("\n");
        for (Language l:lenguajes.values()) sj.add(l.toString());
        return sj.toString();
    }

    Language get(String id) {
        return lenguajes.get(id);
    }

    /**
     * Versión 1 de findMostSimilars. (almacena datos en un array de objetos de una clase interna sin relación de orden)
     * Porporciona un array ordenado con los <code>n</code> lenguajes más próximos al identificado por <code>id</code>
     * donde el primero es el más próximo y el último el más alejado.
     * Si se solicitan más lenguajes de los disponibles, se obtienen todos los disponibles. 
     * 
     * @param id identificador del lenguaje de referencia.
     * @param n el número de lenguajes a obtener.
     * @return el array con los <code>n</code> lenguajes más similares al identificado por <code>id</code>.
     */
    public Language[] findMostSimilars1(String id, int n){
        //control de parámetros (un n excesivo se reduce al máximo posible y no se genera excepción)
        if (lenguajes.get(id)==null) throw new IllegalArgumentException("El lenguaje "+id+" no está disponible.");
        if (n<=0) throw new IllegalArgumentException(n+" no es un número de lenguajes válido.");
        n=Math.min(n, lenguajes.size()-1);
        
        //Generamos una clase local para mantener la relación identificador_de_idioma-similaridad_con_el_de_referencia 
        class X{
            String id; double similarity; 
            public X(String id, double similarity) {
                this.id = id;
                this.similarity = similarity;
            }
        }
        
        //Guardaremos todas los objetos X (identificador-similaridad) en una lista
        List<X> xList=new ArrayList<>();
        Language lang=lenguajes.get(id);
        for (String id0:lenguajes.keySet()) xList.add(new X(id0, Language.similarity(lang, lenguajes.get(id0))));

        //Pasamos de la lista a un array y lo ordenamos usando un comparador
        X[] xArray=new X[xList.size()];
        xList.toArray(xArray);
        Arrays.sort(xArray,new Comparator<X>(){
            @Override
            public int compare(X o1, X o2) {
                return -Double.compare(o1.similarity, o2.similarity);
            }
        });
        
        //Extraemos del array ordenado de objetos X los n primeros lenguajes evitando el primero
        Language[] ret=new Language[n];
        for (int i=0; i<n; i++) {ret[i]=lenguajes.get(xArray[i+1].id);}
        return ret;
    }
    
    /**
     * Versión 2 de findMostSimilars. (almacena datos en un array de objetos de una clase interna con relación de orden)
     * Porporciona un array ordenado con los <code>n</code> lenguajes más próximos al identificado por <code>id</code>
     * donde el primero es el más próximo y el último el más alejado.
     * Si se solicitan más lenguajes de los disponibles, se obtienen todos los disponibles. 
     * 
     * @param id identificador del lenguaje de referencia.
     * @param n el número de lenguajes a obtener.
     * @return el array con los <code>n</code> lenguajes más similares al identificado por <code>id</code>.
     */
    public Language[] findMostSimilars2(String id, int n){
        //control de parámetros (un n excesivo se reduce al máximo posible y no se genera excepción)
        if (lenguajes.get(id)==null) throw new IllegalArgumentException("El lenguaje "+id+" no está disponible.");
        if (n<=0) throw new IllegalArgumentException(n+" no es un número de lenguajes válido.");
        n=Math.min(n, lenguajes.size()-1);
        
        //Generamos una clase local para mantener la relación identificador_de_idioma-similaridad_con_el_de_referencia
        //Hacemos que los objetos X sean comparables
        class X implements Comparable<X>{
            String id; double similarity; 
            public X(String id, double similarity) {
                this.id = id;
                this.similarity = similarity;
            }

            @Override
            public int compareTo(X o) {
                return -Double.compare(similarity, o.similarity);
            }
        }

        //Guardaremos todas los objetos X (identificador-similaridad) en una lista
        List<X> xList=new ArrayList<>();
        Language lang=lenguajes.get(id);
        for (String id0:lenguajes.keySet()) xList.add(new X(id0, Language.similarity(lang, lenguajes.get(id0))));

        //Pasamos de la lista a un array y lo ordenamos con su orden preestablecido
        X[] xArray = xList.toArray(new X[xList.size()]);
        Arrays.sort(xArray);
        
        //Extraemos del array ordenado de objetos X los n primeros lenguajes evitando el primero
        Language[] ret=new Language[n];
        for (int i=0; i<n; i++) {ret[i]=lenguajes.get(xArray[i+1].id);}
        return ret;
    }

    
    /**
     * Versión 3 de findMostSimilars. (almacena datos en un <code>Set</code> ordenado de objetos de una clase interna con relación de orden)
     * Porporciona un array ordenado con los <code>n</code> lenguajes más próximos al identificado por <code>id</code>
     * donde el primero es el más próximo y el último el más alejado.
     * Si se solicitan más lenguajes de los disponibles, se obtienen todos los disponibles. 
     * 
     * @param id identificador del lenguaje de referencia.
     * @param n el número de lenguajes a obtener.
     * @return el array con los <code>n</code> lenguajes más similares al identificado por <code>id</code>.
     */
    public Language[] findMostSimilars3(String id, int n){
        //control de parámetros (un n excesivo se reduce al máximo posible y no se genera excepción)
        if (lenguajes.get(id)==null) throw new IllegalArgumentException("El lenguaje "+id+" no está disponible.");
        if (n<=0) throw new IllegalArgumentException(n+" no es un número de lenguajes válido.");
        n=Math.min(n, lenguajes.size()-1);

        //Generamos una clase local para mantener la relación identificador_de_idioma-similaridad_con_el_de_referencia
        //Hacemos que los objetos X sean comparables
        class X implements Comparable<X>{
            String id; double similarity; 
            public X(String id, double similarity) {
                this.id = id;
                this.similarity = similarity;
            }

            @Override
            public int compareTo(X o) {
                return -Double.compare(similarity, o.similarity);
            }
        }

        //Guardaremos todas los objetos X (identificador-similaridad) en un conjunto ordenado
        Set<X> xSet=new TreeSet<>();
        Language lang=lenguajes.get(id);
        for (String id0:lenguajes.keySet()) xSet.add(new X(id0, Language.similarity(lang, lenguajes.get(id0))));
        
        //Extraemos del conjunto ordenado de objetos X los n primeros lenguajes evitando el primero
        //Para ello necesitamos un Iterador
        Language[] ret=new Language[n];
        Iterator<X> xIt=xSet.iterator();
        xIt.next();
        for (int i=0; i<n; i++) {ret[i]=lenguajes.get(xIt.next().id);}
        return ret;
    }
    
    /**
     * Versión 4 de findMostSimilars. (almacena datos en un <code>Map</code> ordenado)
     * Porporciona un array ordenado con los <code>n</code> lenguajes más próximos al identificado por <code>id</code>
     * donde el primero es el más próximo y el último el más alejado.
     * Si se solicitan más lenguajes de los disponibles, se obtienen todos los disponibles. 
     * 
     * @param id identificador del lenguaje de referencia.
     * @param n el número de lenguajes a obtener.
     * @return el array con los <code>n</code> lenguajes más similares al identificado por <code>id</code>.
     */
    public Language[] findMostSimilars4(String id, int n){
        //control de parámetros (un n excesivo se reduce al máximo posible y no se genera excepción)
        if (lenguajes.get(id)==null) throw new IllegalArgumentException("El lenguaje "+id+" no está disponible.");
        if (n<=0) throw new IllegalArgumentException(n+" no es un número de lenguajes válido.");
        n=Math.min(n, lenguajes.size()-1);

        //Guardaremos todas las similaridades e identificadores en un arbol ordenado
        //OJO- Ordenamos por 1-similaridad para que el orden disponga primero los más similares
        Map<Double, String> map=new TreeMap<>();
        Language lang=lenguajes.get(id);
        for (String id0:lenguajes.keySet()) map.put(1-Language.similarity(lang, lenguajes.get(id0)),id0);
        
        //Extraemos del conjunto ordenado de objetos X los n primeros lenguajes evitando el primero
        //Para ello necesitamos un Iterador
        Language[] ret=new Language[n];
        Iterator<Double> xIt=map.keySet().iterator();
        xIt.next();
        for (int i=0; i<n; i++) {ret[i]=lenguajes.get(map.get(xIt.next()));}
        return ret;
    }
}